##
# File: UniProtExtractor.py
# Date: 5-Dec-2020  jdw
#
# Utilities to extract selected details from the UniProt exchange collections.
#
##
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.exdb.utils.ObjectExtractor import ObjectExtractor
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class UniProtExtractor(object):
    """Utilities to extract selected details from the UniProt exchange collections."""

    def __init__(self, cfgOb):
        self.__cfgOb = cfgOb

    def exportReferenceSequenceDetails(self, filePath, fmt="json"):
        rD = self.getReferenceSequenceDetails()
        mU = MarshalUtil()
        ok = mU.doExport(filePath, rD, fmt=fmt, indent=3)
        logger.info("Exporting (%d) UniProt reference sequences (status=%r)", len(rD), ok)
        return ok

    def getReferenceSequenceDetails(self):
        """Get reference protein sequence essential details (sequence, taxonomy, name, gene, ...)"""
        uD = None
        try:
            obEx = ObjectExtractor(
                self.__cfgOb,
                databaseName="uniprot_exdb",
                collectionName="reference_entry",
                useCache=False,
                keyAttribute="uniprot",
                uniqueAttributes=["rcsb_id"],
                selectionQuery={},
                selectionList=[
                    "source_scientific",
                    "taxonomy_id",
                    "rcsb_id",
                    "gene",
                    "names",
                    "sequence",
                ],
            )
            #
            eCount = obEx.getCount()
            logger.info("Reference entry count is %d", eCount)
            objD = obEx.getObjects()
            rD = {}
            for rId, uD in objD.items():
                taxId = uD["taxonomy_id"]
                sn = uD["source_scientific"]
                sequence = uD["sequence"]
                gn = None
                pn = None
                if "gene" in uD:
                    for tD in uD["gene"]:
                        if tD["type"] == "primary":
                            gn = tD["name"]
                            break
                for tD in uD["names"]:
                    if tD["nameType"] == "recommendedName":
                        pn = tD["name"]
                        break
                rD[rId] = {"accession": rId, "taxId": taxId, "scientific_name": sn, "gene": gn, "name": pn, "sequence": sequence}

        except Exception as e:
            logger.exception("Failing uD %r with %s", uD, str(e))
        #
        return rD



##
# File:    GlycanProviderTests.py
# Author:  J. Westbrook
# Date:    24-May-2021
#
# Update:
#
#
##
"""
Tests for accessors for managing Glycan extracted annotations.

"""

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import unittest

from rcsb.exdb.branch.GlycanProvider import GlycanProvider
from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class GlycanProviderTests(unittest.TestCase):
    def setUp(self):
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__mU = MarshalUtil(workPath=self.__cachePath)
        #
        self.__mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data")
        self.__configPath = os.path.join(TOPDIR, "rcsb", "mock-data", "config", "dbload-setup-example.yml")
        configName = "site_info_configuration"
        self.__cfgOb = ConfigUtil(configPath=self.__configPath, defaultSectionName=configName, mockTopPath=self.__mockTopPath)
        #
        self.__stashUrl = None
        self.__stashRemotePath = os.path.join(self.__cachePath, "stash-remote")

    def tearDown(self):
        pass

    def testGlycanMapping(self):
        minCount = 12
        gP = GlycanProvider(cachePath=self.__cachePath, useCache=False)
        ok = gP.testCache(minCount=0)
        self.assertTrue(ok)
        ok = gP.update(self.__cfgOb, fmt="json", indent=3)
        self.assertTrue(ok)
        riD = gP.getIdentifiers()
        logger.info("riD (%d)", len(riD))
        ok = gP.testCache(minCount=minCount)
        self.assertTrue(ok)
        gP = GlycanProvider(cachePath=self.__cachePath, useCache=True)
        ok = gP.testCache(minCount=minCount)
        self.assertTrue(ok)

    @unittest.skip("Internal test")
    def testGlycanStashRemote(self):
        minCount = 12
        configName = "site_info_remote_configuration"
        cfgOb = ConfigUtil(configPath=self.__configPath, defaultSectionName=configName, mockTopPath=self.__mockTopPath)
        gP = GlycanProvider(cachePath=self.__cachePath, useCache=True)
        ok = gP.testCache(minCount=minCount)
        ok = gP.update(cfgOb, fmt="json", indent=3)
        self.assertTrue(ok)
        riD = gP.getIdentifiers()
        logger.info("riD (%d)", len(riD))
        self.assertTrue(ok)
        ok = gP.backup(cfgOb, configName)
        self.assertTrue(ok)
        ok = gP.restore(cfgOb, configName)
        self.assertTrue(ok)
        #
        ok = gP.update(cfgOb, fmt="json", indent=3)
        self.assertTrue(ok)
        #


def glycanMappingSuite():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(GlycanProviderTests("testGlycanMapping"))
    return suiteSelect


if __name__ == "__main__":

    mySuite = glycanMappingSuite()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
#  File:           ChemRefMappingProvider.py
#  Date:           18-Jun-2021 jdw
#
#  Updated:
#
##
"""
Accessors for chemical reference identifier mapping data.
"""

import datetime
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.exdb.chemref.ChemRefExtractor import ChemRefExtractor

logger = logging.getLogger(__name__)


class ChemRefMappingProvider(StashableBase):
    """Accessors for chemical reference identifier mapping data."""

    def __init__(self, cachePath, useCache=True):
        #
        self.__cachePath = cachePath
        self.__useCache = useCache
        self.__dirName = "chemref-mapping"
        super(ChemRefMappingProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__rD = {}
        self.__mapD = self.__reload(self.__dirPath, useCache)
        #

    def testCache(self, minCount=0):
        logger.info("Mapping count %d", len(self.__mapD["mapping"]) if "mapping" in self.__mapD else 0)
        if minCount == 0 or self.__mapD and "mapping" in self.__mapD and len(self.__mapD["mapping"]) >= minCount:
            return True
        else:
            return False

    def getReferenceIds(self, referenceResourceName, localId):
        """Get the identifiers in the reference resource corresponding to input local
        identifiers (Chemical Component or BIRD).

        Args:
            referenceResourceName (str): chemical reference resource name (DrugBank, ChEMBL, ChEBI, PubChem, ...)
            localId (str): local identifier for a Chemical Component or BIRD definition

        Returns:
            list: list of reference identifiers
        """
        if not self.__rD:
            for rN, forwardD in self.__mapD["mapping"].items():
                # {refId :[lId, lId, ...], ...}
                reverseD = {}
                for refId, rcsbIdL in forwardD.items():
                    for rId in rcsbIdL:
                        reverseD.setdefault(rId, []).append(refId)
                self.__rD[rN] = reverseD
        #
        try:
            return self.__rD[referenceResourceName.upper()][localId]
        except Exception:
            return []

    def getLocalIds(self, referenceResourceName, referenceId):
        """Get the local identifiers (Chemical Component or BIRD) corresponding to identifiers in
        chemical reference resource.

        Args:
            referenceResourceName (str): chemical reference resource name (DrugBank, ChEMBL, ChEBI, PubChem, ...)
            referenceId (str): identifier in the chemical reference resource

        Returns:
            list: list of local Chemical Component or BIRD identifiers
        """
        try:
            return self.__mapD["mapping"][referenceResourceName.upper()][referenceId]
        except Exception:
            return []

    def __getMappingDataPath(self):
        return os.path.join(self.__dirPath, "chemref-mapping-data.json")

    def __reload(self, dirPath, useCache):
        startTime = time.time()
        fD = {}
        ok = False
        mappingPath = self.__getMappingDataPath()
        #
        logger.info("useCache %r mappingPath %r", useCache, mappingPath)
        if useCache and self.__mU.exists(mappingPath):
            fD = self.__mU.doImport(mappingPath, fmt="json")
            ok = True
        else:
            fU = FileUtil()
            fU.mkdir(dirPath)
        # ---
        logger.info("Completed reload with status (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return fD

    def fetchChemRefMapping(self, cfgOb, referenceResourceNameList=None):
        """Fetch reference resource mapping for chemical component and BIRD definitions

        Args:
            cfgOb (obj): instance configuration class ConfigUtil()
            referenceResourceNameList (list, optional): list of chemical reference resources. Defaults to [DrugBank, ChEMBL].

        Returns:
            bool: True for success or False otherwise
        """
        try:
            rnL = referenceResourceNameList if referenceResourceNameList is not None else ["DrugBank", "ChEMBL"]
            mD = {}
            crExt = ChemRefExtractor(cfgOb)
            for referenceResourceName in rnL:
                idD = crExt.getChemCompAccessionMapping(referenceResourceName=referenceResourceName)
                logger.info("%s mapping dictionary (%d)", referenceResourceName, len(idD))
                mD[referenceResourceName.upper()] = idD
            #
            fp = self.__getMappingDataPath()
            tS = datetime.datetime.now().isoformat()
            vS = datetime.datetime.now().strftime("%Y-%m-%d")
            ok = self.__mU.doExport(fp, {"version": vS, "created": tS, "mapping": mD}, fmt="json", indent=3)
            return ok
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False



##
# File:    PubChemEtlWrapperTests.py
# Author:  J. Westbrook
# Date:    20-Jul-2020
#
# Updates:
#
##
"""
Tests for PubChem ETL wrapper methods
"""

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import platform
import resource
import time
import unittest

from rcsb.exdb.chemref.PubChemEtlWrapper import PubChemEtlWrapper

from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))


class PubChemEtlWrapperTests(unittest.TestCase):
    def __init__(self, methodName="runTest"):
        super(PubChemEtlWrapperTests, self).__init__(methodName)
        self.__verbose = True

    def setUp(self):
        #
        self.__workPath = os.path.join(HERE, "test-output")
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        self.__dirPath = os.path.join(self.__cachePath, "PubChem")
        self.__mU = MarshalUtil(workPath=self.__cachePath)
        #
        # Site configuration used for database resource access -
        self.__mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data")
        configPath = os.path.join(TOPDIR, "rcsb", "mock-data", "config", "dbload-setup-example.yml")
        configName = "site_info_configuration"
        self.__cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=configName, mockTopPath=self.__mockTopPath)
        #
        # These are test source files for chemical component/BIRD indices
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        self.__numComponents = 30
        self.__numSelectMatches = 23
        self.__numAltMatches = 2
        self.__numTotalMatches = 50
        #
        self.__startTime = time.time()
        logger.debug("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testAFromBootstrap(self):
        """Test case - build CCD/BIRD search indices and search for PubChem matches."""
        try:
            #  -- Update local chemical indices and  create PubChem mapping index ---

            pcewP = PubChemEtlWrapper(self.__cfgOb, self.__cachePath)
            ok = pcewP.updateIndex(
                ccUrlTarget=self.__ccUrlTarget,
                birdUrlTarget=self.__birdUrlTarget,
                ccFileNamePrefix="cc-abbrev",
                exportPath=self.__dirPath,
                rebuildChemIndices=True,
                numProc=4,
            )
            self.assertTrue(ok)
            #
            mL = pcewP.getMatches()
            self.assertGreaterEqual(len(mL), self.__numTotalMatches)
            selectMatchD, altMatchD = pcewP.getSelectedMatches()
            #
            logger.info("matchD (%d)", len(selectMatchD))
            self.assertGreaterEqual(len(selectMatchD), self.__numSelectMatches)
            self.assertGreaterEqual(len(altMatchD), self.__numAltMatches)
            #
            ok = pcewP.dump(contentType="index")
            self.assertTrue(ok)
            ok = pcewP.toStash(contentType="index")
            self.assertTrue(ok)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()

    def testBFromRestore(self):
        """Test case - operations from a restored starting point"""
        try:
            #  --
            pcewP = PubChemEtlWrapper(self.__cfgOb, self.__cachePath)
            ok = pcewP.fromStash(contentType="index")
            self.assertTrue(ok)
            #
            numObjects = pcewP.reloadDump(contentType="index")
            logger.info("Restored %d correspondence records", numObjects)
            self.assertGreaterEqual(numObjects, self.__numComponents)
            mapD, extraMapD = pcewP.getSelectedMatches(exportPath=os.path.join(self.__cachePath, "mapping"))
            self.assertGreaterEqual(len(mapD), self.__numSelectMatches)
            logger.info("mapD (%d) extraMapD (%d) %r", len(mapD), len(extraMapD), extraMapD)
            self.assertGreaterEqual(len(extraMapD), self.__numAltMatches)
            cidList = pcewP.getMatches()
            logger.info("cidList (%d)", len(cidList))
            self.assertGreaterEqual(len(cidList), self.__numTotalMatches - 2)
            ok = pcewP.updateMatchedData()
            self.assertTrue(ok)
            ok = pcewP.dump(contentType="data")
            self.assertTrue(ok)
            ok = pcewP.toStash(contentType="data")
            self.assertTrue(ok)
            ok = pcewP.updateIdentifiers()
            self.assertTrue(ok)
            ok = pcewP.dump(contentType="identifiers")
            self.assertTrue(ok)
            ok = pcewP.toStash(contentType="identifiers")
            self.assertTrue(ok)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()


def pubChemEtlWrapperSuite():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(PubChemEtlWrapperTests("testAFromBootstrap"))
    suiteSelect.addTest(PubChemEtlWrapperTests("testBFromRestore"))
    return suiteSelect


if __name__ == "__main__":
    mySuite = pubChemEtlWrapperSuite()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    PubChemDataCacheProviderTests.py
# Author:  J. Westbrook
# Date:    17-Jul-2020
#
# Updates:
#
##
"""
Tests for reference data cache maintenance operations
"""

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import platform
import resource
import time
import unittest

from rcsb.exdb.chemref.PubChemDataCacheProvider import PubChemDataCacheProvider

from rcsb.utils.config.ConfigUtil import ConfigUtil

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))


class PubChemDataCacheProviderTests(unittest.TestCase):
    def __init__(self, methodName="runTest"):
        super(PubChemDataCacheProviderTests, self).__init__(methodName)
        self._verbose = True

    def setUp(self):
        #
        self.__cidList = ["49866376", "66835630", "71664579", "11915", "12072107"]
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        #
        # Site configuration used for database resource access -
        self.__mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data")
        configPath = os.path.join(TOPDIR, "rcsb", "mock-data", "config", "dbload-setup-example.yml")
        configName = "site_info_configuration"
        self.__cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=configName, mockTopPath=self.__mockTopPath)
        #
        self.__startTime = time.time()
        logger.debug("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testALoadAndUpdate(self):
        """Test case - load and reload/update data store."""
        try:
            #  -- Update/create cache ---
            exportPath = os.path.join(self.__cachePath, "PubChem")
            pcdcP = PubChemDataCacheProvider(self.__cfgOb, self.__cachePath)
            ok, failList = pcdcP.load(self.__cidList, exportPath=exportPath)
            self.assertTrue(ok)
            self.assertEqual(len(failList), 0)
            logger.info("Status %r failList %r", ok, failList)
            #
            idL = pcdcP.getRefIdCodes()
            logger.info("idL %r", idL)
            self.assertGreaterEqual(len(idL), len(self.__cidList))
            #
            ok, failList = pcdcP.updateMissing(self.__cidList, exportPath=exportPath)
            self.assertTrue(ok)
            self.assertEqual(len(failList), 0)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()

    def testBackupAndRestore(self):
        """Test case - load and dump operations."""
        try:
            #  -- Backup/Restore cache ---
            pcdcP = PubChemDataCacheProvider(self.__cfgOb, self.__cachePath)
            ok, failList = pcdcP.load(self.__cidList, exportPath=None)
            self.assertEqual(len(failList), 0)
            self.assertTrue(ok)
            ok = pcdcP.dump(fmt="json")
            self.assertTrue(ok)
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()

    def testGetRelatedIdentifiers(self):
        """Test case - get PubChem xrefs."""
        try:
            #  --- Get related identifiers ---
            pcdcP = PubChemDataCacheProvider(self.__cfgOb, self.__cachePath)
            rD = pcdcP.getRelatedMapping(self.__cidList)
            logger.info("rD %r", rD)
            self.assertGreaterEqual(len(rD), len(self.__cidList))
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()


def pubChemDataCacheProviderSuite():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(PubChemDataCacheProviderTests("testALoadAndUpdate"))
    suiteSelect.addTest(PubChemDataCacheProviderTests("testBackupAndRestore"))
    suiteSelect.addTest(PubChemDataCacheProviderTests("testGetRelatedIdentifiers"))

    return suiteSelect


if __name__ == "__main__":
    mySuite = pubChemDataCacheProviderSuite()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    PubChemIndexCacheProviderTests.py
# Author:  J. Westbrook
# Date:    16-Jul-2020
#
# Updates:
#
##
"""
Tests for PubChem index cache maintenance operations
"""

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import platform
import resource
import time
import unittest

from rcsb.exdb.chemref.PubChemIndexCacheProvider import PubChemIndexCacheProvider

from rcsb.utils.config.ConfigUtil import ConfigUtil

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))


class PubChemIndexCacheProviderTests(unittest.TestCase):
    def __init__(self, methodName="runTest"):
        super(PubChemIndexCacheProviderTests, self).__init__(methodName)
        self.__verbose = True

    def setUp(self):
        #
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        #
        # Site configuration used for database resource access -
        self.__mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data")
        configPath = os.path.join(TOPDIR, "rcsb", "mock-data", "config", "dbload-setup-example.yml")
        self.__configName = "site_info_configuration"
        self.__cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=self.__configName, mockTopPath=self.__mockTopPath)
        #
        # These are test source files for chemical component/BIRD indices
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        self.__startTime = time.time()
        logger.debug("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testAPubChemIndexCacheProvider(self):
        """Test case - search, backup, restore and select PubChem correspondences for reference chemical definitions."""
        try:
            #  -- Update/create mapping index cache ---
            numObj = 30
            pcicP = PubChemIndexCacheProvider(self.__cfgOb, self.__cachePath)
            pcicP.updateMissing(
                expireDays=0,
                cachePath=self.__cachePath,
                ccUrlTarget=self.__ccUrlTarget,
                birdUrlTarget=self.__birdUrlTarget,
                ccFileNamePrefix="cc-abbrev",
                exportPath=os.path.join(self.__cachePath, "PubChem"),
                rebuildChemIndices=False,
                fetchLimit=None,
            )
            matchD = pcicP.getMatchData(expireDays=0)
            logger.info("matchD (%d)", len(matchD))
            self.assertGreaterEqual(len(matchD), numObj)
            ok = pcicP.testCache()
            self.assertTrue(ok)
            #
            ok = pcicP.dump()
            self.assertTrue(ok)
            #
            mapD, extraMapD = pcicP.getSelectedMatches(exportPath=os.path.join(self.__cachePath, "mapping"))
            self.assertGreaterEqual(len(mapD), 20)
            logger.info("mapD (%d) extraMapD (%d) %r", len(mapD), len(extraMapD), extraMapD)
            self.assertGreaterEqual(len(extraMapD), 2)
            cidList = pcicP.getMatches()
            logger.info("cidList (%d)", len(cidList))
            self.assertGreaterEqual(len(cidList), 49)
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()

    def testBPubChemIndexCacheProvider(self):
        """Test case -  verify the PubChem index cache"""
        try:
            #  -- check cache
            pcicP = PubChemIndexCacheProvider(self.__cfgOb, self.__cachePath)
            ok = pcicP.testCache()
            self.assertTrue(ok)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()


def pubChemIndexCacheProviderSuite():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(PubChemIndexCacheProviderTests("testAPubChemIndexCacheProvider"))
    suiteSelect.addTest(PubChemIndexCacheProviderTests("testBPubChemIndexCacheProviderCache"))
    return suiteSelect


if __name__ == "__main__":
    mySuite = pubChemIndexCacheProviderSuite()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File: BranchedEntityExtractor.py
# Date: 24-May-2021  jdw
#
# Utilities to extract selected details from the core branched entity collections.
#
#
# Updates:
#
##
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.exdb.utils.ObjectExtractor import ObjectExtractor
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class BranchedEntityExtractor(object):
    """Utilities to extract selected details from the core branched entity collections."""

    def __init__(self, cfgOb):
        self.__cfgOb = cfgOb

    def exportBranchedEntityDetails(self, filePath, fmt="json"):
        """Export branched entity details (BIRD mapping and WURCS descriptors)"""
        rD = self.getBranchedDetails()
        # ----
        mU = MarshalUtil()
        ok = mU.doExport(filePath, rD, fmt=fmt, indent=3)
        logger.info("Exporting (%d) branched entities status %r", len(rD), ok)
        return ok

    def getBranchedDetails(self):
        """Get branched entity details (BIRD mapping and WURCS descriptors)"""
        rD = {}
        try:

            #
            obEx = ObjectExtractor(
                self.__cfgOb,
                databaseName="pdbx_core",
                collectionName="pdbx_core_branched_entity",
                useCache=False,
                keyAttribute="entity",
                uniqueAttributes=["rcsb_id"],
                selectionQuery={},
                selectionList=["rcsb_id", "pdbx_entity_branch_descriptor", "rcsb_branched_entity_container_identifiers"],
            )
            #
            # eCount = obEx.getCount()
            # logger.info("Branched entity count is %d", eCount)
            objD = obEx.getObjects()
            rD = {}
            for _, eD in objD.items():
                rcsbId = eD["rcsb_id"]
                #
                prdId = None
                try:
                    pD = eD["rcsb_branched_entity_container_identifiers"]
                    prdId = pD["prd_id"]
                except Exception:
                    pass
                #
                wurcs = None
                try:
                    for tD in eD["pdbx_entity_branch_descriptor"]:
                        if tD["type"] == "WURCS":
                            wurcs = tD["descriptor"]
                except Exception:
                    pass
                if prdId or wurcs:
                    rD[rcsbId] = {"prdId": prdId, "wurcs": wurcs}

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return rD



##
# File:    GlycanUtilsTests.py
# Author:  J. Westbrook
# Date:    25-May-2021
#
# Update:
##
"""
Tests for creating glycan accession mapping details.
"""

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import unittest

from rcsb.exdb.branch.GlycanUtils import GlycanUtils
from rcsb.utils.config.ConfigUtil import ConfigUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(HERE))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class GlycanUtilsTests(unittest.TestCase):
    def setUp(self):
        self.__dirPath = os.path.join(HERE, "test-output", "CACHE", "glycan")
        #
        self.__mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data")
        configPath = os.path.join(TOPDIR, "rcsb", "mock-data", "config", "dbload-setup-example.yml")
        configName = "site_info_configuration"
        self.__cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=configName, mockTopPath=self.__mockTopPath)
        #

    def tearDown(self):
        pass

    def testGlycanMapping(self):
        gU = GlycanUtils(self.__cfgOb, self.__dirPath)
        beD = gU.getBranchedEntityDetails()
        self.assertGreaterEqual(len(beD), 1)
        logger.info("branched entity descriptor details (%d)", len(beD))
        eaD = gU.updateEntityAccessionMap()
        logger.info("updated entity accession map length (%d)", len(eaD))
        self.assertGreaterEqual(len(eaD), 12)


def glycanMappingSuite():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(GlycanUtilsTests("testGlycanMapping"))
    return suiteSelect


if __name__ == "__main__":

    mySuite = glycanMappingSuite()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File: PubChemIndexCacheProvider.py
# Date: 2-Apr-2020  jdw
#
# Utilities to manage chemical component/BIRD to PubChem compound identifier mapping data.
#
# Updates:
#  9-May-2020 jdw separate cache behavior with separate option rebuildChemIndices=True/False
# 16-Jul-2020 jdw separate index and reference data management.
# 23-Jul-2021 jdw Make PubChemIndexCacheProvider a subclass of StashableBase()
#
##
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import time

from rcsb.exdb.utils.ObjectExtractor import ObjectExtractor
from rcsb.exdb.utils.ObjectUpdater import ObjectUpdater
from rcsb.utils.chem.ChemCompIndexProvider import ChemCompIndexProvider
from rcsb.utils.chem.ChemCompSearchIndexProvider import ChemCompSearchIndexProvider
from rcsb.utils.chemref.PubChemUtils import PubChemUtils, ChemicalIdentifier
from rcsb.utils.io.IoUtil import getObjSize
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.io.TimeUtil import TimeUtil
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil


logger = logging.getLogger(__name__)


class PubChemUpdateWorker(object):
    """A skeleton worker class that implements the interface expected by the multiprocessing module
    for fetching CCD/BIRD to PubChem chemical compound identifier correspondences --
    """

    def __init__(self, cfgOb, searchIdxD, **kwargs):
        self.__cfgOb = cfgOb
        self.__searchIdxD = searchIdxD
        #
        _ = kwargs
        self.__lookupD = {}
        for sId, sD in self.__searchIdxD.items():
            ccId = sId.split("|")[0]
            self.__lookupD.setdefault(ccId, []).append(sD)
        self.__databaseName = "pubchem_exdb"
        self.__matchIndexCollectionName = "reference_match_index"
        self.__createCollections(self.__databaseName, self.__matchIndexCollectionName, indexAttributeNames=["rcsb_id", "rcsb_last_update"])
        self.__pcU = PubChemUtils()

    def __genChemIdList(self, ccId):
        """Return a list of ChemicalIdentifier() objects for the input chemical component identifier.

        Args:
            ccId (str): chemical component identifiers

        Returns:
            (list): list of ChemicalIdentifier() objects corresponding to the input chemical component.
        """
        chemIdList = []
        if ccId in self.__lookupD:
            for sD in self.__lookupD[ccId]:
                if "inchi-key" in sD:
                    idType = "inchikey"
                    descr = sD["inchi-key"]
                elif "smiles" in sD:
                    idType = "smiles"
                    descr = sD["smiles"]
                chemIdList.append(ChemicalIdentifier(idCode=ccId, identifierSource=sD["build-type"], identifierType=idType, identifier=descr, indexName=sD["name"]))
        return chemIdList

    def updateList(self, dataList, procName, optionsD, workingDir):
        """Update the input list of reference data identifiers (ChemicalIdentifier()) and return
        matching diagnostics and reference feature data.
        {
               "_id" : ObjectId("5e8dfb49eab967a0483a0472"),
               "rcsb_id" : "local reference ID (ccid|bird)", << LOCAL CANONICAL ID (e.g. ATP, PRD_000100)
               "rcsb_last_update" : ISODate("2020-04-08T16:26:47.993+0000"),
               "matched_ids" : [
                   {"matched_id":  "<external reference ID code>", "search_id_type" : "oe-smiles", "search_id_source": "model-xyz",
                                   'source_index_name': <>, 'source_inchikey': <>, 'source_smiles': <>},
                   {"matched_id":  "<external reference ID code>", "search_id_type": ... , "search_id_source": ... , ...}
                   ]                          ]
               },
           }
           // Failed matches are recorded with NO matchedIds:
           {
               "_id" : ObjectId("5e8dfb49eab967a0483a04a3"),
               "rcsb_id" : "local reference ID (ccid|bird)", << LOCAL ID
               "rcsb_last_update" : ISODate("2020-04-08T16:26:48.025+0000"),
           }
           #
        """
        _ = workingDir
        chunkSize = optionsD.get("chunkSize", 50)
        matchIdOnly = optionsD.get("matchIdOnly", True)
        # Path to store raw request data -
        exportPath = optionsD.get("exportPath", None)
        #
        successList = []
        retList1 = []
        retList2 = []
        diagList = []
        emptyList = []
        #
        try:
            tU = TimeUtil()
            ccIdList = dataList
            numChunks = len(list(self.__chunker(ccIdList, chunkSize)))
            logger.info("%s search starting for %d reference definitions (in chunks of length %d)", procName, len(ccIdList), chunkSize)
            for ii, ccIdChunk in enumerate(self.__chunker(ccIdList, chunkSize), 1):
                logger.info("%s starting chunk for %d of %d", procName, ii, numChunks)
                # tDL = []
                tIdxDL = []
                timeS = tU.getDateTimeObj(tU.getTimestamp())
                for ccId in ccIdChunk:
                    # Get various forms from the search index -
                    chemIdList = self.__genChemIdList(ccId)
                    tIdxD = {"rcsb_id": ccId, "rcsb_last_update": timeS}
                    #
                    mL = []
                    for chemId in chemIdList:
                        stA = time.time()
                        ok, refDL = self.__pcU.assemble(chemId, exportPath=exportPath, matchIdOnly=matchIdOnly)
                        #
                        if not ok:
                            etA = time.time()
                            logger.debug("Failing %s search source %s for %s (%.4f secs)", chemId.identifierType, chemId.identifierSource, chemId.idCode, etA - stA)

                        #
                        if ok and refDL:
                            for tD in refDL:
                                pcId = tD["cid"]
                                inchiKey = (
                                    self.__searchIdxD[chemId.indexName]["inchi-key"]
                                    if chemId.indexName in self.__searchIdxD and "inchi-key" in self.__searchIdxD[chemId.indexName]
                                    else None
                                )
                                smiles = (
                                    self.__searchIdxD[chemId.indexName]["smiles"] if chemId.indexName in self.__searchIdxD and "smiles" in self.__searchIdxD[chemId.indexName] else None
                                )
                                mL.append(
                                    {
                                        "matched_id": pcId,
                                        "search_id_type": chemId.identifierType,
                                        "search_id_source": chemId.identifierSource,
                                        "source_index_name": chemId.indexName,
                                        "source_smiles": smiles,
                                        "source_inchikey": inchiKey,
                                    }
                                )
                                # tD.update({"rcsb_id": pcId, "rcsb_last_update": timeS})
                                # tDL.append(tD)
                    #
                    if mL:
                        tIdxD["matched_ids"] = mL
                        successList.append(ccId)
                    else:
                        logger.info("No match result for any form of %s", ccId)
                    #
                    tIdxDL.append(tIdxD)
                # --
                startTimeL = time.time()
                logger.info("Saving chunk %d (len=%d)", ii, len(ccIdChunk))
                self.__updateObjectStore(self.__databaseName, self.__matchIndexCollectionName, tIdxDL)
                endTimeL = time.time()
                logger.info("Saved chunk %d (len=%d) in %.3f secs", ii, len(ccIdChunk), endTimeL - startTimeL)
        except Exception as e:
            logger.exception("Failing %s for %d data items %s", procName, len(dataList), str(e))
        logger.info("%s dataList length %d success length %d rst1 %d rst2 %d", procName, len(dataList), len(successList), len(retList1), len(retList2))
        #
        return successList, emptyList, emptyList, diagList

    def __updateObjectStore(self, databaseName, collectionName, objDL):
        updateDL = []
        for objD in objDL:
            try:
                selectD = {"rcsb_id": objD["rcsb_id"]}
                updateDL.append({"selectD": selectD, "updateD": objD})
            except Exception as e:
                logger.exception("Failing with %s", str(e))
        obUpd = ObjectUpdater(self.__cfgOb)
        numUpd = obUpd.update(databaseName, collectionName, updateDL)
        logger.info("Updated reference count is %d", numUpd)

    def __createCollections(self, databaseName, collectionName, indexAttributeNames=None):
        obUpd = ObjectUpdater(self.__cfgOb)
        ok = obUpd.createCollection(databaseName, collectionName, indexAttributeNames=indexAttributeNames, checkExists=True, bsonSchema=None)
        return ok

    def __chunker(self, iList, chunkSize):
        chunkSize = max(1, chunkSize)
        return (iList[i : i + chunkSize] for i in range(0, len(iList), chunkSize))


class PubChemIndexCacheProvider(StashableBase):
    """Utilities to manage chemical component/BIRD to PubChem compound identifier mapping data."""

    def __init__(self, cfgOb, cachePath):
        dirName = "PubChem-index"
        super(PubChemIndexCacheProvider, self).__init__(cachePath, [dirName])
        self.__cfgOb = cfgOb
        self.__cachePath = cachePath
        self.__dirPath = os.path.join(self.__cachePath, dirName)
        #
        self.__databaseName = "pubchem_exdb"
        self.__matchIndexCollectionName = "reference_match_index"
        #

        self.__matchD = None

    def getMatchData(self, expireDays=0):
        if not self.__matchD:
            selectD = {}
            if expireDays > 0:
                tU = TimeUtil()
                tS = tU.getTimestamp(useUtc=True, before={"days": expireDays})
                selectD.update({"rcsb_latest_update": {"$lt": tU.getDateTimeObj(tS)}})
            self.__matchD = self.__getReferenceData(self.__databaseName, self.__matchIndexCollectionName, selectD=selectD)
            #
        return self.__matchD

    def testCache(self, minMatch=None, logSizes=False):
        self.getMatchData()
        okC = bool(self.__matchD)
        if not okC:
            return okC
        logger.info("Reference data cache lengths: matchD %d", len(self.__matchD))
        if minMatch and len(self.__matchD) < minMatch:
            return False
        #
        if logSizes:
            logger.info("PubChem MatchD %.2f", getObjSize(self.__matchD) / 1000000.0)
        return True

    def __getdumpFilePath(self, fmt="json"):
        stashBaseFileName = "pubchem_match_index_object_list"
        fExt = ".json" if fmt == "json" else ".pic"
        fp = os.path.join(self.__dirPath, stashBaseFileName + fExt)
        return fp

    def dump(self, fmt="json"):
        """Dump PubChem index reference data from the object store.

        Args:
            fmt (str, optional): [description]. Defaults to "json".

        Returns:
            bool: True for success or False otherwise
        """
        ok = False
        try:
            self.getMatchData()
            if fmt in ["json", "pickle"]:
                kwargs = {}
                fp = self.__getdumpFilePath(fmt=fmt)
                logger.info("Saving object store to %s", fp)
                mU = MarshalUtil(workPath=self.__dirPath)
                if fmt in ["json"]:
                    kwargs = {"indent": 3}
                ok = mU.doExport(fp, self.__matchD, fmt=fmt, **kwargs)
        except Exception as e:
            logger.exception("Failing for %r with %s", self.__dirPath, str(e))
        return ok

    def reloadDump(self, fmt="json"):
        """Reload PubChem reference data store from saved dump.

        Args:
            fmt (str, optional): format of the backup file (pickle or json). Defaults to "json".

        Returns:
            (int): number of objects restored.
        """
        numUpd = 0
        try:
            # Read from disk backup and update object store -
            if fmt in ["json", "pickle"]:
                fp = self.__getdumpFilePath(fmt="json")
                logger.info("Restoring object store from %s", fp)
                mU = MarshalUtil(workPath=self.__dirPath)
                matchD = mU.doImport(fp, fmt=fmt)
                numUpd = self.__reloadDump(matchD, self.__databaseName, self.__matchIndexCollectionName, indexAttributeNames=["rcsb_id", "rcsb_last_update"])
        except Exception as e:
            logger.exception("Failing for %r with %s", self.__dirPath, str(e))
        # --
        return numUpd

    def updateMissing(self, expireDays=0, fetchLimit=None, updateUnmatched=True, numProc=12, **kwargs):
        """Update match index from object store

        Args:
            expireDays (int): expiration days on match data (default 0 meaning none)
            fetchLimit (int): limit to the number of entry updates performed (None)
            updateUnmatched (bool): Previously unmatched search definitions will be retried on update (default=True)
            numProc (int): for rebuilding local chemical indices the number processors to apply (default=12)

        Returns:
            bool: True for success or False otherwise

        ChemicalIdentifierFields = ("idCode", "identifierSource", "identifierType", "identifier")
        ChemicalIdentifier = collections.namedtuple("ChemicalIdentifier", ChemicalIdentifierFields, defaults=(None,) * len(ChemicalIdentifierFields))


                // Failed matches are recorded with NO matchedIds:
                {
                    "_id" : ObjectId("5e8dfb49eab967a0483a04a3"),
                    "rcsb_id" : "local reference ID (ccid|bird)", << LOCAL ID
                    "rcsb_last_update" : ISODate("2020-04-08T16:26:48.025+0000"),
                }
        """
        #
        matchD = {}
        matchedIdList = []
        try:
            # ---
            # Get current the indices of source chemical reference data -
            ok, ccidxP, ccsidxP = self.__rebuildChemCompSourceIndices(numProc, **kwargs)
            if not ok:
                return matchD
            #
            ccIdxD = ccidxP.getIndex()
            searchIdxD = ccsidxP.getIndex()
            # Index of target of local chemical component and BIRD identifiers
            sourceIdList = sorted(ccIdxD.keys())
            logger.info("Reloading chemical reference data (expireDays %r, updateUnmatched %r)", expireDays, updateUnmatched)
            matchedIdList = self.__getMatchIndexIds(searchIdxD, expireDays=expireDays, updateUnmatched=updateUnmatched)
            # --
            logger.info("Starting matched reference identifier count (%d) ", len(matchedIdList))
            updateIdList = sorted(set(sourceIdList) - set(matchedIdList))
            logger.info("Missing chemical definition correspondences %d fetchLimit %r", len(updateIdList), fetchLimit)
            #
            updateIdList = updateIdList[:fetchLimit] if fetchLimit else updateIdList
            #
            if updateIdList:
                logger.info("Update reference data cache for %d chemical identifers", len(updateIdList))
                ok, failList = self.__updateReferenceData(updateIdList, searchIdxD, **kwargs)
                logger.info("Update reference data return status is %r missing count %d", ok, len(failList))
            else:
                logger.info("No reference data updates required")
            # --
            return ok
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def getMatches(self):
        """Get all PubChem correspondences from the current match index..

        Returns:

            (list): PubChem compound identifier codes.

        """
        self.getMatchData()
        #
        pcidList = []
        try:
            pcidS = set()
            for _, mD in self.__matchD.items():
                if "matched_ids" in mD:
                    for sD in mD["matched_ids"]:
                        pcidS.add(sD["matched_id"])
            pcidList = list(pcidS)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return pcidList

    def getSelectedMatches(self, **kwargs):
        """Select preferred PubChem correspondences from the current match index for the input source component build type.
            and separatel return alternative matches for other source types.

        Args:
            sourceTypes (list, optional):  list of source chemical component build types (default: ["model-xyz"])
            exportPath: (str, optional): export path for correspondences

        Returns:
            dict, dict : mapD { ccId1: [{'pcId': ... , 'inchiKey': ... }], ccId2: ...},
                         altD { ccId1: [{'pcId': ... , 'inchiKey': ... 'sourceType': ... }], ccId2: ...}

                Example match index entry:
                {
                    "_id" : ObjectId("5e8dfb49eab967a0483a0472"),
                    "rcsb_id" : "local reference ID (ccid|bird)", << LOCAL CANONICAL ID (e.g. ATP, PRD_000100)
                    "rcsb_last_update" : ISODate("2020-04-08T16:26:47.993+0000"),
                    "matched_ids" : [
                        {"matched_id":  "<external reference ID code>", "search_id_type" : "oe-smiles", "search_id_source": "model-xyz",
                                        'source_index_name': <>, 'source_inchikey': <>, 'source_smiles': <>},
                        {"matched_id":  "<external reference ID code>", "search_id_type": ... , "search_id_source": ... , ...}
                        ]                          ]
                    },
                }
        """
        #
        self.getMatchData()

        sourceTypes = kwargs.get("sourceTypes", ["model-xyz"])
        exportPath = kwargs.get("exportPath", None)
        #
        mapD = {}
        altMapD = {}
        extraMapD = {}
        try:
            for ccId, mD in self.__matchD.items():
                if "matched_ids" in mD:
                    for sD in mD["matched_ids"]:
                        #
                        if sD and "search_id_source" in sD:
                            pcId = sD["matched_id"]
                            inchiKey = sD["source_inchikey"]
                            #
                            if sD["search_id_source"] in sourceTypes:
                                mapD.setdefault(ccId, []).append({"pcId": pcId, "inchiKey": inchiKey})
                            else:
                                altMapD.setdefault(ccId, []).append({"pcId": pcId, "inchiKey": inchiKey, "sourceType": sD["search_id_source"]})
            #
            difS = set(altMapD.keys()) - set(mapD.keys())
            logger.info("PubChem preferred correspondence length (%d) alternative extras (%d)", len(mapD), len(difS))
            for ccId in difS:
                extraMapD[ccId] = altMapD[ccId]
            if exportPath:
                fp = os.path.join(exportPath, "pubchem_matches.json")
                mU = MarshalUtil(workPath=exportPath)
                mU.doExport(fp, mapD, fmt="json", indent=3)
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return mapD, extraMapD

    #
    # -- Extract current data from object store --
    def __getMatchIndexIds(self, searchIdxD, expireDays=0, updateUnmatched=True):
        """Get CCD/BIRD reference data identifiers in the current match index subject to an
           expiration interval (i.e. not matched or older than deltaDays).

        Args:
            searchIdxD (dict): CCD/BIRD search index dictionary
            expireDays (int, optional): expiration interval in days. Defaults to 0 (no expiration).
            updateUnmatched (bool, optional): include only matched identifiers (i.e. exclude any tried but unmatched cases)

        Returns:
            (list): chemical component/BIRD reference identifier list
        """
        selectD = {}
        if expireDays > 0:
            tU = TimeUtil()
            tS = tU.getTimestamp(useUtc=True, before={"days": expireDays})
            selectD.update({"rcsb_latest_update": {"$lt": tU.getDateTimeObj(tS)}})
        #
        if updateUnmatched:
            # Return only cases with an existing correspondence
            selectD.update({"matched_ids": {"$exists": True}})
        matchD = self.__getReferenceData(self.__databaseName, self.__matchIndexCollectionName, selectD=selectD if selectD else None)
        #
        # For the selected cases in the index-
        retIdList = []
        if searchIdxD:
            # Exclude definitions if source InChIKey in the match index differs with the Key in the current search index.
            for ccId, inD in matchD.items():
                if updateUnmatched and "matched_ids" not in inD:
                    retIdList.append(ccId)
                    continue
                hasChanged = False
                for mD in inD["matched_ids"]:
                    if mD["source_index_name"] not in searchIdxD:
                        hasChanged = True
                        logger.info("Identifier %s no longer in search index", mD["source_index_name"])
                        break
                    if mD["source_inchikey"] != searchIdxD[mD["source_index_name"]]["inchi-key"]:
                        logger.info("Identifier %s InChIKey changed search index", mD["source_index_name"])
                        hasChanged = True
                        break
                if not hasChanged:
                    retIdList.append(ccId)
        #
        return sorted(retIdList)

    #
    def __getReferenceData(self, databaseName, collectionName, selectD=None, selectionList=None):
        logger.info("Searching %s %s with selection query %r", databaseName, collectionName, selectD)
        obEx = ObjectExtractor(
            self.__cfgOb,
            databaseName=databaseName,
            collectionName=collectionName,
            keyAttribute="rcsb_id",
            uniqueAttributes=["rcsb_id"],
            selectionQuery=selectD,
            selectionList=selectionList,
            stripObjectId=True,
        )
        docCount = obEx.getCount()
        logger.info("Reference data object count %d", docCount)
        objD = obEx.getObjects()
        return objD

    def __updateReferenceData(self, idList, searchIdxD, **kwargs):
        """Launch worker methods to update chemical reference data correspondences.

        Args:
            idList (list): list of local chemical identifiers (ChemIdentifier())

        Returns:
            (bool, list): status flag, list of unmatched identifiers
        """
        numProc = 1
        chunkSize = 50
        exportPath = kwargs.get("exportPath", None)
        logger.info("Length starting list is %d", len(idList))
        optD = {"chunkSize": chunkSize, "exportPath": exportPath, "matchIdOnly": True}
        rWorker = PubChemUpdateWorker(self.__cfgOb, searchIdxD)
        if numProc > 1:
            mpu = MultiProcUtil(verbose=True)
            mpu.setOptions(optD)
            mpu.set(workerObj=rWorker, workerMethod="updateList")
            ok, failList, resultList, _ = mpu.runMulti(dataList=idList, numProc=numProc, numResults=2, chunkSize=chunkSize)
            logger.info("Multi-proc %r failures %r result lengths %r %r", ok, len(failList), len(resultList[0]), len(resultList[1]))
        else:
            successList, _, _, _ = rWorker.updateList(idList, "SingleProc", optD, self.__dirPath)
            failList = list(set(idList) - set(successList))
            ok = len(failList) > 0
            logger.info("Single-proc status %r failures %r", ok, len(failList))
        #
        return ok, failList

    def __reloadDump(self, objD, databaseName, collectionName, indexAttributeNames=None):
        """Internal method to restore the input database/collection using the input data object.

        Args:
            objD (obj): Target reference or index data object
            databaseName (str): target database name
            collectionName (str): target collection name
            indexAttributeNames (list, optional): Primary index attributes. Defaults to None.

        Returns:
            int: inserted or updated object count
        """
        try:
            numUpd = 0
            numTotal = 0
            updateDL = []
            for entityKey, obj in objD.items():
                if "_id" in obj:
                    obj.pop("_id")
                selectD = {"rcsb_id": entityKey}
                updateDL.append({"selectD": selectD, "updateD": obj})
            #
            obUpd = ObjectUpdater(self.__cfgOb)
            ok = obUpd.createCollection(databaseName, collectionName, indexAttributeNames=indexAttributeNames, checkExists=True, bsonSchema=None)
            if ok:
                numUpd = obUpd.update(databaseName, collectionName, updateDL)
                logger.debug("Updated object count is %d", numUpd)
            else:
                logger.error("Create %s %s failed", databaseName, collectionName)
            numTotal = obUpd.count(databaseName, collectionName)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return numTotal

    #                           --- --- ---
    # -- Load or rebuild source chemical reference data indices --
    def __rebuildChemCompSourceIndices(self, numProc, **kwargs):
        """Rebuild source indices of chemical component definitions."""
        logger.info("Rebuilding chemical definition index.")
        ok1, ccidxP = self.__buildChemCompIndex(**kwargs)
        logger.info("Rebuilding chemical search indices.")
        ok2, ccsidxP = self.__buildChemCompSearchIndex(numProc, **kwargs)
        return ok1 & ok2, ccidxP, ccsidxP

    def __buildChemCompIndex(self, **kwargs):
        """Build chemical component cache files from the input component dictionaries"""
        try:
            molLimit = kwargs.get("molLimit", None)
            useCache = not kwargs.get("rebuildChemIndices", False)
            logSizes = kwargs.get("logSizes", False)
            ccFileNamePrefix = kwargs.get("ccFileNamePrefix", "cc-full")
            ccUrlTarget = kwargs.get("ccUrlTarget", None)
            birdUrlTarget = kwargs.get("birdUrlTarget", None)
            cachePath = kwargs.get("cachePath", self.__cachePath)
            #
            ccidxP = ChemCompIndexProvider(
                ccUrlTarget=ccUrlTarget, birdUrlTarget=birdUrlTarget, cachePath=cachePath, useCache=useCache, molLimit=molLimit, ccFileNamePrefix=ccFileNamePrefix
            )
            ok = ccidxP.testCache(minCount=molLimit, logSizes=logSizes)
            return ok, ccidxP if ok else None
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return False, None

    def __buildChemCompSearchIndex(self, numProc, **kwargs):
        """Test build search index chemical component cache files from the input component dictionaries"""
        try:
            cachePath = kwargs.get("cachePath", self.__cachePath)
            molLimit = kwargs.get("molLimit", None)
            useCache = not kwargs.get("rebuildChemIndices", False)
            logSizes = kwargs.get("logSizes", False)
            limitPerceptions = kwargs.get("limitPerceptions", False)
            #
            # numProc = kwargs.get("numProc", 1)
            # numProc = self.__numProc
            chunkSize = kwargs.get("chunkSize", 5)
            molLimit = kwargs.get("molLimit", None)
            ccFileNamePrefix = kwargs.get("ccFileNamePrefix", "cc-full")
            quietFlag = kwargs.get("quietFlag", True)
            ccUrlTarget = kwargs.get("ccUrlTarget", None)
            birdUrlTarget = kwargs.get("birdUrlTarget", None)
            #
            ccsiP = ChemCompSearchIndexProvider(
                ccUrlTarget=ccUrlTarget,
                birdUrlTarget=birdUrlTarget,
                cachePath=cachePath,
                useCache=useCache,
                molLimit=molLimit,
                ccFileNamePrefix=ccFileNamePrefix,
                limitPerceptions=limitPerceptions,
                numProc=numProc,
                maxChunkSize=chunkSize,
                quietFlag=quietFlag,
            )
            ok = ccsiP.testCache(minCount=molLimit, logSizes=logSizes)
            return ok, ccsiP if ok else None
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False, None



##
# File:    PdbxLoaderFixture.py
# Author:  J. Westbrook
# Date:    4-Sep-2019
# Version: 0.001
#
# Updates:
#
##
"""
Fixture for loading the chemical reference and pdbx_core collections in a loca mongo instance.

"""

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import glob
import logging
import os
import platform
import resource
import time
import unittest

from rcsb.db.mongo.DocumentLoader import DocumentLoader
from rcsb.db.mongo.PdbxLoader import PdbxLoader
from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.io.FileUtil import FileUtil

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()
logger.setLevel(logging.INFO)

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))


class PdbxLoaderFixture(unittest.TestCase):

    def __init__(self, methodName="runTest"):
        super(PdbxLoaderFixture, self).__init__(methodName)
        self.__verbose = True

    def setUp(self):
        #
        #
        self.__isMac = platform.system() == "Darwin"
        self.__excludeType = None if self.__isMac else "optional"
        self.__mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data")
        configPath = os.path.join(TOPDIR, "rcsb", "mock-data", "config", "dbload-setup-example.yml")
        # configPath = os.path.join(TOPDIR, "rcsb", "mock-data", "config", "dbload-setup-example-local.yml")
        # To Do: Investigate why GitUtil sometimes gives divergence error when using 'DISCOVERY_MODE: remote', but not with 'local':
        #            stderr: 'fatal: Need to specify how to reconcile divergent branches.'
        #        Behavior isn't entirely predictable, since it happens sometimes but not all the time.
        #        To fully debug, will need to add more logging statements to GitUtil, StashableBase, & StashUtil (in rcsb.utils.io)
        #        Or, can try to resolve error directly by specifying how to reconcile diverent branches in git.Repo class.
        configName = "site_info_configuration"
        self.__cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=configName, mockTopPath=self.__mockTopPath)
        #
        self.__resourceName = "MONGO_DB"
        self.__failedFilePath = os.path.join(HERE, "test-output", "failed-list.txt")
        self.__cachePath = os.path.join(TOPDIR, "CACHE")
        self.__readBackCheck = True
        self.__numProc = 2
        self.__chunkSize = 10
        self.__fileLimit = 38
        self.__documentStyle = "rowwise_by_name_with_cardinality"
        #
        self.__birdChemCompCoreIdList = [
            "PRD_000010",
            "PRD_000060",
            "PRD_000220",
            "PRD_000882",
            "PRD_000154",
            "PRD_000877",
            "PRD_000198",
            "PRD_000009",
            "PRD_000979",
            "PRDCC_000010",
            "PRDCC_000220",
            "PRDCC_000882",
            "PRDCC_000154",
            "PRDCC_000198",
            "PRDCC_000009",
            "FAM_000010",
            "FAM_000210",
            "FAM_000220",
            "FAM_000001",
            "FAM_000391",
            "FAM_000093",
            "FAM_000084",
            "FAM_000016",
            "FAM_000336",
            "1G1",
            "2RT",
            "2XL",
            "2XN",
            "ATP",
            "BJA",
            "BM3",
            "CNC",
            "DAL",
            "DDZ",
            "DHA",
            "DSN",
            "GTP",
            "HKL",
            "NAC",
            "NAG",
            "NND",
            "PTR",
            "SEP",
            "SMJ",
            "STL",
            "UNK",
            "UNX",
            "UVL",
        ]
        #
        self.__pdbIdList = [
            "1ah1",
            "1b5f",
            "1bmv",
            "1c58",
            "1dsr",
            "1dul",
            "1kqe",
            "1o3q",
            "1sfo",
            "2hw3",
            "2hyv",
            "2osl",
            "2voo",
            "2wmg",
            "3ad7",
            "3hya",
            "3iyd",
            "3mbg",
            "3rer",
            "3vd8",
            "3vfj",
            "3x11",
            "3ztj",
            "4e2o",
            "4en8",
            "4mey",
            "5eu8",
            "5kds",
            "5tm0",
            "5vh4",
            "5vp2",
            "6fsz",
            "6lu7",
            "6nn7",
            "6q20",
            "6rfk",
            "6rku",
            "6yrq",
        ]
        self.__ldList = [
            {
                "databaseName": "bird_chem_comp_core",
                "collectionNameList": None,
                "loadType": "full",
                "mergeContentTypes": None,
                "validationLevel": "min",
                "inputIdCodeList": self.__birdChemCompCoreIdList
            },
            {
                "databaseName": "pdbx_core",
                "collectionNameList": None,
                "loadType": "full",
                "mergeContentTypes": ["vrpt"],
                "validationLevel": "min",
                "inputIdCodeList": self.__pdbIdList
            },
            {
                "databaseName": "pdbx_comp_model_core",
                "collectionNameList": None,
                "loadType": "full",
                "mergeContentTypes": None,
                "validationLevel": "min",
                "inputIdCodeList": None
            },
        ]
        #
        self.__modelFixture()
        self.__startTime = time.time()
        logger.debug("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def __modelFixture(self):
        fU = FileUtil()
        modelSourcePath = os.path.join(self.__mockTopPath, "AF")
        for iPath in glob.iglob(os.path.join(modelSourcePath, "*.cif.gz")):
            fn = os.path.basename(iPath)
            uId = fn.split("-")[1]
            h3 = uId[-2:]
            h2 = uId[-4:-2]
            h1 = uId[-6:-4]
            oPath = os.path.join(self.__cachePath, "computed-models", h1, h2, h3, fn)
            fU.put(iPath, oPath)

    def testPdbxLoader(self):
        #
        for ld in self.__ldList:
            self.__pdbxLoaderWrapper(**ld)

    def __pdbxLoaderWrapper(self, **kwargs):
        """Wrapper for the PDBx loader module"""
        try:
            logger.info("Loading %s", kwargs["databaseName"])
            mw = PdbxLoader(
                self.__cfgOb,
                cachePath=self.__cachePath,
                resourceName=self.__resourceName,
                numProc=self.__numProc,
                chunkSize=self.__chunkSize,
                fileLimit=kwargs.get("fileLimit", self.__fileLimit),
                verbose=self.__verbose,
                readBackCheck=self.__readBackCheck,
                maxStepLength=2000,
                useSchemaCache=True,
                rebuildSchemaFlag=False,
            )
            ok = mw.load(
                kwargs["databaseName"],
                collectionLoadList=kwargs["collectionNameList"],
                loadType=kwargs["loadType"],
                inputPathList=None,
                inputIdCodeList=kwargs["inputIdCodeList"],
                styleType=self.__documentStyle,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=self.__failedFilePath,
                saveInputFileListPath=None,
                pruneDocumentSize=None,
                logSize=False,
                validationLevel=kwargs["validationLevel"],
                mergeContentTypes=kwargs["mergeContentTypes"],
                useNameFlag=False,
                providerTypeExclude=self.__excludeType,
                restoreUseGit=True,
                restoreUseStash=False,
            )
            self.assertTrue(ok)
            ok = self.__loadStatus(mw.getLoadStatus())
            self.assertTrue(ok)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()

    def __loadStatus(self, statusList):
        sectionName = "data_exchange_configuration"
        dl = DocumentLoader(
            self.__cfgOb,
            self.__cachePath,
            resourceName=self.__resourceName,
            numProc=self.__numProc,
            chunkSize=self.__chunkSize,
            documentLimit=None,
            verbose=self.__verbose,
            readBackCheck=self.__readBackCheck,
        )
        #
        databaseName = self.__cfgOb.get("DATABASE_NAME", sectionName=sectionName)
        collectionName = self.__cfgOb.get("COLLECTION_UPDATE_STATUS", sectionName=sectionName)
        ok = dl.load(databaseName, collectionName, loadType="append", documentList=statusList, indexAttributeList=["update_id", "database_name", "object_name"], keyNames=None)
        return ok


def mongoLoadPdbxSuite():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(PdbxLoaderFixture("testPdbxLoader"))
    return suiteSelect


if __name__ == "__main__":
    mySuite = mongoLoadPdbxSuite()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    TreeNodeListWorkerTests.py
# Author:  J. Westbrook
# Date:    23-Apr-2019
#
# Updates:
#
##
"""
Tests for for tree node list worker ---

"""

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
import os
import platform
import resource
import time
import unittest

from rcsb.db.utils.TimeUtil import TimeUtil
from rcsb.exdb.tree.TreeNodeListWorker import TreeNodeListWorker
from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))


class TreeNodeListWorkerTests(unittest.TestCase):
    def __init__(self, methodName="runTest"):
        super(TreeNodeListWorkerTests, self).__init__(methodName)
        self.__verbose = True

    def setUp(self):
        self.__isMac = platform.system() == "Darwin"
        self.__doLoad = True if self.__isMac else False
        #
        #
        self.__mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data")
        configPath = os.path.join(TOPDIR, "rcsb", "mock-data", "config", "dbload-setup-example.yml")
        #
        configName = "site_info_configuration"
        self.__cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=configName, mockTopPath=self.__mockTopPath)
        #
        self.__cachePath = os.path.join(TOPDIR, "CACHE")
        #
        self.__mU = MarshalUtil()
        #
        self.__readBackCheck = True
        self.__numProc = 2
        self.__chunkSize = 10
        self.__documentLimit = None
        self.__debugFlag = False
        self.__loadType = "full"
        self.__useCache = True
        #
        self.__startTime = time.time()
        logger.debug("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testTreeLoader(self):
        """Test case - extract entity polymer info"""
        try:
            tU = TimeUtil()
            updateId = tU.getCurrentWeekSignature()
            rhw = TreeNodeListWorker(
                self.__cfgOb,
                self.__cachePath,
                numProc=self.__numProc,
                chunkSize=self.__chunkSize,
                documentLimit=self.__documentLimit,
                verbose=self.__debugFlag,
                readBackCheck=self.__readBackCheck,
                useCache=self.__useCache,
            )
            #
            ok = rhw.load(updateId, loadType=self.__loadType, doLoad=self.__doLoad)
            self.assertTrue(ok)
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()


def treeNodeListSuite():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(TreeNodeListWorkerTests("testTreeLoader"))
    return suiteSelect


if __name__ == "__main__":
    mySuite = treeNodeListSuite()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    EntryInfoProviderTests.py
# Author:  J. Westbrook
# Date:    22-Sep-2021
#
# Update:
#
#
##
"""
Tests for accessors entry-level annotations.

"""

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import unittest

from rcsb.exdb.entry.EntryInfoProvider import EntryInfoProvider
from rcsb.utils.config.ConfigUtil import ConfigUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class EntryInfoProviderTests(unittest.TestCase):
    doInternal = False

    def setUp(self):
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        #
        self.__mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data")
        self.__configPath = os.path.join(TOPDIR, "rcsb", "mock-data", "config", "dbload-setup-example.yml")
        configName = "site_info_configuration"
        self.__cfgOb = ConfigUtil(configPath=self.__configPath, defaultSectionName=configName, mockTopPath=self.__mockTopPath)
        #

    def tearDown(self):
        pass

    def testGetEntryInfo(self):
        minCount = 12
        eiP = EntryInfoProvider(cachePath=self.__cachePath, useCache=False)
        ok = eiP.testCache(minCount=0)
        self.assertTrue(ok)
        ok = eiP.update(self.__cfgOb, fmt="json", indent=3)
        self.assertTrue(ok)
        riD = eiP.getEntryInfo("4en8")
        logger.info("riD (%d) %r", len(riD), riD)
        rL = eiP.getEntriesByPolymerEntityCount(count=2)
        self.assertGreaterEqual(len(rL), 5)
        ok = eiP.testCache(minCount=minCount)
        self.assertTrue(ok)
        eiP = EntryInfoProvider(cachePath=self.__cachePath, useCache=True)
        ok = eiP.testCache(minCount=minCount)
        self.assertTrue(ok)

    @unittest.skipUnless(doInternal, "Internal full test")
    def testEntryInfoRemote(self):
        minCount = 182000
        configName = "site_info_remote_configuration"
        cfgOb = ConfigUtil(configPath=self.__configPath, defaultSectionName=configName, mockTopPath=self.__mockTopPath)
        eiP = EntryInfoProvider(cachePath=self.__cachePath, useCache=True)
        ok = eiP.testCache(minCount=minCount)
        ok = eiP.update(cfgOb, fmt="json", indent=3)
        self.assertTrue(ok)
        riD = eiP.getEntryInfo("1kip")
        logger.info("riD (%d)", len(riD))
        self.assertGreaterEqual(len(riD), 1)
        configName = "site_info_configuration"
        cfgOb = ConfigUtil(configPath=self.__configPath, defaultSectionName=configName, mockTopPath=self.__mockTopPath)
        ok = eiP.backup(cfgOb, configName, useStash=False, useGit=True)
        self.assertTrue(ok)
        ok = eiP.restore(cfgOb, configName, useStash=False, useGit=True)
        self.assertTrue(ok)
        eiP = EntryInfoProvider(cachePath=self.__cachePath, useCache=True)
        ok = eiP.testCache(minCount=minCount)


def entryInfoSuite():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(EntryInfoProviderTests("testGetEntryInfo"))
    return suiteSelect


if __name__ == "__main__":

    mySuite = entryInfoSuite()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



# File: setup.py
# Date: 23-Feb-2019
#
# Updates:
#
#
import re

from setuptools import find_packages
from setuptools import setup

packages = []
thisPackage = "rcsb.exdb"

with open("rcsb/exdb/cli/__init__.py", "r", encoding="utf-8") as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE).group(1)


# Load packages from requirements*.txt
with open("requirements.txt", "r", encoding="utf-8") as ifh:
    packagesRequired = [ln.strip() for ln in ifh.readlines()]

with open("README.md", "r", encoding="utf-8") as ifh:
    longDescription = ifh.read()

if not version:
    raise RuntimeError("Cannot find version information")

setup(
    name=thisPackage,
    version=version,
    description="RCSB Python ExDB data extraction and loading workflows",
    long_description_content_type="text/markdown",
    long_description=longDescription,
    author="John Westbrook",
    author_email="john.westbrook@rcsb.org",
    url="https://github.com/rcsb/py-rcsb_exdb",
    #
    license="Apache 2.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    entry_points={"console_scripts": ["exdb_exec_cli=rcsb.exdb.cli.ExDbExec:main"]},
    #
    install_requires=packagesRequired[1:],
    packages=find_packages(exclude=["rcsb.mock-data", "rcsb.exdb.tests-anal", "rcsb.exdb.tests-*", "tests.*"]),
    package_data={
        # If any package contains *.md or *.rst ...  files, include them:
        "": ["*.md", "*.rst", "*.txt", "*.cfg"]
    },
    #
    test_suite="rcsb.exdb.tests",
    tests_require=["tox"],
    #
    # Not configured ...
    extras_require={"dev": ["check-manifest"], "test": ["coverage"]},
    # Added for
    command_options={"build_sphinx": {"project": ("setup.py", thisPackage), "version": ("setup.py", version), "release": ("setup.py", version)}},
    # This setting for namespace package support -
    zip_safe=False,
)



##
# File: ObjectValidator.py
# Date: 17-Oct-2019  jdw
#
# Utilities to extract and update object from the document object server including validation.
#
# Updates:
#
##
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from jsonschema import Draft4Validator
from jsonschema import FormatChecker

from rcsb.db.mongo.Connection import Connection
from rcsb.db.mongo.MongoDbUtil import MongoDbUtil
from rcsb.db.processors.DataExchangeStatus import DataExchangeStatus
from rcsb.db.utils.SchemaProvider import SchemaProvider
from rcsb.db.utils.TimeUtil import TimeUtil

logger = logging.getLogger(__name__)


class ObjectValidator(object):
    """Utilities to extract and update object from the document object server with validation."""

    def __init__(self, cfgOb, objectAdapter=None, cachePath=".", useCache=True, **kwargs):
        self.__cfgOb = cfgOb
        self.__oAdapt = objectAdapter
        self.__resourceName = "MONGO_DB"
        _ = kwargs
        self.__statusList = []
        self.__schP = SchemaProvider(self.__cfgOb, cachePath, useCache=useCache)
        self.__valInst = None

    def __getValidator(self, databaseName, collectionName, schemaLevel="full"):
        _ = self.__schP.makeSchemaDef(databaseName, dataTyping="ANY", saveSchema=True)
        cD = self.__schP.makeSchema(databaseName, collectionName, encodingType="JSON", level=schemaLevel, saveSchema=True)
        # Raises exceptions for schema compliance.
        Draft4Validator.check_schema(cD)
        valInst = Draft4Validator(cD, format_checker=FormatChecker())
        return valInst

    def __validateObj(self, databaseName, collectionName, rObj, label=""):
        try:
            eCount = 0
            tId = rObj["rcsb_id"] if rObj and "rcsb_id" in rObj else "anonymous"
            for error in sorted(self.__valInst.iter_errors(rObj), key=str):
                logger.info("Database %s collection %s (%s %r) path %s error: %s", databaseName, collectionName, label, tId, error.path, error.message)
                logger.debug(">>> Failing object is %r", rObj)
                eCount += 1
        except Exception as e:
            logger.exception("Validation failing %s", str(e))

        return eCount

    def doTransform(self, **kwargs):
        desp = DataExchangeStatus()
        statusStartTimestamp = desp.setStartTime()
        #
        databaseName = kwargs.get("databaseName", "pdbx_core")
        collectionName = kwargs.get("collectionName", "pdbx_core_entry")
        selectionQueryD = kwargs.get("selectionQuery", {})
        fetchLimit = kwargs.get("fetchLimit", None)
        #

        #
        tU = TimeUtil()
        updateId = kwargs.get("updateId", tU.getCurrentWeekSignature())
        #
        docSelectList = self.__selectObjectIds(databaseName, collectionName, selectionQueryD)
        docSelectList = docSelectList[:fetchLimit] if fetchLimit else docSelectList

        ok = self.__transform(databaseName, collectionName, docSelectList)
        #
        if updateId:
            okS = self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)
        return ok and okS

    def __selectObjectIds(self, databaseName, collectionName, selectionQueryD):
        """Return a list of object identifiers for the input selection query."""
        try:

            with Connection(cfgOb=self.__cfgOb, resourceName=self.__resourceName) as client:
                mg = MongoDbUtil(client)
                if mg.collectionExists(databaseName, collectionName):
                    logger.info("%s %s document count is %d", databaseName, collectionName, mg.count(databaseName, collectionName))
                    qD = {}
                    if selectionQueryD:
                        qD.update(selectionQueryD)
                    selectL = ["_id"]
                    dL = mg.fetch(databaseName, collectionName, selectL, queryD=qD)
                    logger.info("Selection %r fetch result count %d", selectL, len(dL))

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return dL
        #

    def __transform(self, databaseName, collectionName, docSelectList, logIncrement=100):
        """Return a list of object identifiers for the input selection query."""
        #
        ok = True
        try:
            self.__valInst = self.__getValidator(databaseName, collectionName, schemaLevel="full")
            with Connection(cfgOb=self.__cfgOb, resourceName=self.__resourceName) as client:
                mg = MongoDbUtil(client)
                if mg.collectionExists(databaseName, collectionName):
                    numDoc = len(docSelectList)
                    for ii, dD in enumerate(docSelectList, 1):
                        if "_id" not in dD:
                            continue
                        rObj = mg.fetchOne(databaseName, collectionName, "_id", dD["_id"])
                        del rObj["_id"]
                        #
                        fOk = True

                        if self.__oAdapt:
                            self.__validateObj(databaseName, collectionName, rObj, label="Original")
                            fOk, rObj = self.__oAdapt.filter(rObj)
                            self.__validateObj(databaseName, collectionName, rObj, label="Updated")
                        if fOk:
                            rOk = mg.replace(databaseName, collectionName, rObj, dD)
                            if rOk is None:
                                tId = rObj["rcsb_id"] if rObj and "rcsb_id" in rObj else "anonymous"
                                logger.error("%r %r (%r) failing", databaseName, collectionName, tId)
                                # logger.info("rObj.keys() %r", list(rObj.keys()))
                                # logger.info("rObj.items() %s", rObj.items())
                                rOk = False
                            ok = ok and rOk
                        #
                        if ii % logIncrement == 0 or ii == numDoc:
                            logger.info("Replace status %r object (%d of %d)", ok, ii, numDoc)
                        #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def getLoadStatus(self):
        return self.__statusList

    def __updateStatus(self, updateId, databaseName, collectionName, status, startTimestamp):
        try:
            sFlag = "Y" if status else "N"
            desp = DataExchangeStatus()
            desp.setStartTime(tS=startTimestamp)
            desp.setObject(databaseName, collectionName)
            desp.setStatus(updateId=updateId, successFlag=sFlag)
            desp.setEndTime()
            self.__statusList.append(desp.getStatus())
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False



##
# File: PolymerEntityExtractor.py
# Date: 5-Dec-2020  jdw
#
# Utilities to extract selected details from the core polymer entity collections.
#
#
# Updates:
#
##
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from rcsb.exdb.seq.UniProtExtractor import UniProtExtractor
from rcsb.exdb.utils.ObjectExtractor import ObjectExtractor
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


def getRangeOverlap(entityBeg, entityEnd, refBeg, refEnd):
    r1 = range(entityBeg, entityEnd)
    r2 = range(refBeg, refEnd)
    if r1.start == r1.stop or r2.start == r2.stop:
        return set()
    if not ((r1.start < r2.stop and r1.stop > r2.start) or (r1.stop > r2.start and r2.stop > r1.start)):
        return set()
    return set(range(max(r1.start, r2.start), min(r1.stop, r2.stop) + 1))


class PolymerEntityExtractor(object):
    """Utilities to extract selected details from the core polymer entity collections."""

    def __init__(self, cfgOb):
        self.__cfgOb = cfgOb

    def exportProteinSequenceDetails(self, filePath, fmt="json", minSeqLen=0):
        """Export protein sequence and taxonomy data (required to build protein sequence fasta file)"""
        rD, missingSrcD = self.getProteinSequenceDetails(minSeqLen=minSeqLen)
        # ----
        mU = MarshalUtil()
        ok1 = mU.doExport(filePath, rD, fmt=fmt, indent=3)
        #
        pth, _ = os.path.split(filePath)
        mU = MarshalUtil()
        ok2 = mU.doExport(os.path.join(pth, "missingSrcNames.json"), missingSrcD, fmt="json")
        logger.info("Exporting (%d) protein sequence records with missing source count (%d) status %r", len(rD), len(missingSrcD), ok1 and ok2)

    def getProteinSequenceDetails(self, minSeqLen=0):
        """Get protein sequence and taxonomy data (required to build protein sequence fasta file)"""
        missingSrcD = {}
        rD = {}
        try:
            unpEx = UniProtExtractor(self.__cfgOb)
            unpD = unpEx.getReferenceSequenceDetails()
            #
            obEx = ObjectExtractor(
                self.__cfgOb,
                databaseName="pdbx_core",
                collectionName="pdbx_core_polymer_entity",
                useCache=False,
                keyAttribute="entity",
                uniqueAttributes=["rcsb_id"],
                selectionQuery={"entity_poly.rcsb_entity_polymer_type": "Protein"},
                selectionList=[
                    "rcsb_id",
                    "rcsb_entity_source_organism",
                    "rcsb_polymer_entity.rcsb_source_part_count",
                    "rcsb_polymer_entity.rcsb_source_taxonomy_count",
                    "rcsb_polymer_entity.src_method",
                    "entity_poly",
                    "rcsb_polymer_entity_align",
                ],
            )
            #
            eCount = obEx.getCount()
            logger.info("Polymer entity count is %d", eCount)
            objD = obEx.getObjects()
            rD = {}
            for rId, eD in objD.items():

                try:
                    pD = eD["entity_poly"]
                    seqS = pD["pdbx_seq_one_letter_code_can"]
                    seqLen = len(seqS)
                except Exception:
                    logger.warning("%s no one-letter-code sequence", rId)
                #
                if seqLen < minSeqLen:
                    continue
                #
                srcMethod = None
                try:
                    pD = eD["rcsb_polymer_entity"]
                    srcMethod = pD["src_method"]
                except Exception:
                    pass
                #
                if "rcsb_entity_source_organism" not in eD:
                    logger.debug("%s No source information (%r) skipping (seqLen %d)", rId, srcMethod, seqLen)
                    continue
                try:
                    sL = []
                    for tD in eD["rcsb_entity_source_organism"]:
                        srcName = tD["scientific_name"] if "scientific_name" in tD else None
                        if "beg_seq_num" in tD and "end_seq_num" in tD:
                            begSeqNum = tD["beg_seq_num"]
                            endSeqNum = tD["end_seq_num"] if tD["end_seq_num"] <= seqLen else seqLen
                        else:
                            begSeqNum = 1
                            endSeqNum = seqLen
                        srcId = tD["pdbx_src_id"]
                        srcType = tD["source_type"]
                        taxId = tD["ncbi_taxonomy_id"] if "ncbi_taxonomy_id" in tD else -1
                        if srcName and taxId == -1:
                            missingSrcD.setdefault(srcName, []).append(rId)
                        orgName = tD["ncbi_scientific_name"] if "ncbi_scientific_name" in tD else ""
                        sL.append({"srcId": srcId, "taxId": taxId, "orgName": orgName, "entitySeqBeg": begSeqNum, "entitySeqEnd": endSeqNum})
                    if len(sL) == 1:
                        sL[0]["entitySeqBeg"] = 1
                        sL[0]["entitySeqEnd"] = seqLen

                except Exception as e:
                    logger.exception("Failing for (%r) tD %r with %s", rId, tD, str(e))
                #
                try:
                    pD = eD["rcsb_polymer_entity"]
                    partCount = pD["rcsb_source_part_count"]
                except Exception:
                    logger.warning("%s no source part count", rId)
                    partCount = 1
                try:
                    pD = eD["rcsb_polymer_entity"]
                    taxCount = pD["rcsb_source_taxonomy_count"]
                except Exception:
                    if srcType == "synthetic":
                        taxCount = 0
                    else:
                        logger.warning("%s (srcName %r) no source taxonomy count type %r", rId, srcName, srcType)
                        if srcName:
                            taxCount = 1
                        else:
                            taxCount = 0
                #
                uDL = []
                try:
                    for tD in eD["rcsb_polymer_entity_align"]:
                        uD = {}
                        if tD["reference_database_name"] in ["UniProt", "GenBank", "PIR", "EMBL", "NORINE", "PRF"]:
                            uD["refDbId"] = tD["reference_database_accession"]
                            uD["refDbName"] = tD["reference_database_name"]
                            uD["provSource"] = tD["provenance_source"]
                            if tD["reference_database_accession"] in unpD:
                                uD.update(unpD[tD["reference_database_accession"]])
                            aL = []
                            for qD in tD["aligned_regions"]:
                                if qD["entity_beg_seq_id"] + qD["length"] - 1 > seqLen:
                                    qD["length"] = seqLen - qD["entity_beg_seq_id"] + 1
                                srcId = self.__getSourcePart(rId, sL, qD["entity_beg_seq_id"], qD["length"])

                                aL.append({"srcId": srcId, "entitySeqBeg": qD["entity_beg_seq_id"], "refSeqBeg": qD["ref_beg_seq_id"], "length": qD["length"]})
                            uD["alignList"] = aL
                            uDL.append(uD)
                        else:
                            logger.info("%s reference database %s", rId, tD["reference_database_name"])

                except Exception:
                    pass
                rD[rId] = {"alignmentL": uDL, "sourceOrgL": sL, "partCount": partCount, "taxCount": taxCount, "sequence": seqS, "seqLen": seqLen}

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return rD, missingSrcD

    def __getSourcePart(self, entityId, sourceOrgL, entityBeg, seqLen):
        """Return the source part containing the input entity range -

        Args:
            sourceOrgL (list): list of source dictionaries
            entityBeg (int):  begining entity sequence position (matched region)
            seqLen (int):  length sequence range (matched region)

        Returns:
            (int): corresponding source part id or None
        """
        entityEnd = entityBeg + seqLen - 1
        for sD in sourceOrgL:
            srcId = sD["srcId"]
            if sD["entitySeqBeg"] <= entityBeg and sD["entitySeqEnd"] >= entityEnd:
                return srcId
        #
        if len(sourceOrgL) == 1:
            logger.error("%r (%d) Inconsistent range for beg %r end %r sourceOrgL %r", entityId, len(sourceOrgL), entityBeg, entityEnd, sourceOrgL)
            return 1
        else:
            ovTupL = []
            for sD in sourceOrgL:
                srcId = sD["srcId"]
                logger.debug("%r %r beg %r end %r beg %r end %r", entityId, srcId, sD["entitySeqBeg"], sD["entitySeqEnd"], entityBeg, entityEnd)
                oVS = getRangeOverlap(sD["entitySeqBeg"], sD["entitySeqEnd"], entityBeg, entityEnd)
                ovTupL.append((srcId, len(oVS)))
            rL = sorted(ovTupL, key=lambda x: x[1], reverse=True)
            logger.debug("ovTupL %r", rL)
            #
            return rL[0][0]

    def exportProteinEntityFasta(self, fastaPath, taxonPath, detailsPath, minSeqLen=10):
        """Export protein entity Fasta file and associated taxon mapping file (for mmseqs2)

        Args:
            fastaPath (str): protein sequence FASTA output file path
            taxonPath (str): taxon mapping file path (seqid TaxId) (tdd format)
            detailPath (str): protein entity details file path (json)

        Returns:
            bool: True for success or False otherwise

        Example:
            "5H7D_1": {
                    "alignmentL": [
                        {
                            "refDbId": "P42588",
                            "refDbName": "UniProt",
                            "provSource": "PDB",
                            "accession": "P42588",
                            "taxId": 83333,
                            "scientific_name": "Escherichia coli (strain K12)",
                            "gene": "patA",
                            "name": "PATase",
                            "alignList": [
                            {
                                "srcId": "1",
                                "entitySeqBeg": 5,
                                "refSeqBeg": 7,
                                "length": 447
                            }
                            ]
                        },
                        {
                            "refDbId": "P38507",
                            "refDbName": "UniProt",
                            "provSource": "PDB",
                            "accession": "P38507",
                            "taxId": 1280,
                            "scientific_name": "Staphylococcus aureus",
                            "gene": "spa",
                            "name": "IgG-binding protein A",
                            "alignList": [
                            {
                                "srcId": "2",
                                "entitySeqBeg": 452,
                                entitySeqBeg"220,
                                "length": 48
                            }
                            ]
                        }
                    ],
                    "sourceOrgL": [
                        {
                            "srcId": "1",
                            "taxId": 83333,
                            "orgName": "Escherichia coli K-12",
                            "entitySeqBeg": 1,
                            "entitySeqEnd": 451
                        },
                        {
                            "srcId": "2",
                            "taxId": 1280,
                            "orgName": "Staphylococcus aureus",
                            "entitySeqBeg": 452,
                            "entitySeqEnd": 499
                        }
                    ],
                    "partCount": 2,
                    "taxCount": 2,
                    "sequence": "GSHMSASALACSAHALNLIEKRTLDHEEMKALNREVIEYFKEHVNPGF...",
                    "seqLen": 499
                },
        >1ABC_#|prt|<taxid>|beg|end|refdb|refId|refTaxId|refbeg|refend|ref_gn|ref_nm
        """
        proteinSeqD, _ = self.getProteinSequenceDetails(minSeqLen=minSeqLen)
        ok = False

        try:
            taxonL = []
            seqDict = {}
            for eId, eD in proteinSeqD.items():
                #
                seq = eD["sequence"]
                for sD in eD["sourceOrgL"]:
                    srcId = sD["srcId"]
                    taxId = sD["taxId"]
                    seqBeg = int(sD["entitySeqBeg"])
                    seqEnd = int(sD["entitySeqEnd"])
                    seqLen = 1 + (seqEnd - seqBeg)
                    # orgName = sD["orgName"]
                    cD = {"sequence": seq[seqBeg - 1 : seqEnd], "entityId": eId, "srcId": srcId, "seqBeg": seqBeg, "seqEnd": seqEnd, "seqLen": seqLen, "taxId": taxId}
                    seqId = ""
                    cL = []
                    for k, v in cD.items():
                        if k in ["sequence"]:
                            continue
                        cL.append(str(v))
                        cL.append(str(k))
                    seqId = "|".join(cL)
                    seqDict[seqId] = cD
                    taxonL.append("%s\t%s" % (seqId, taxId))
                # ----
            mU = MarshalUtil()
            ok = mU.doExport(detailsPath, proteinSeqD, fmt="json", indent=3)
            ok = mU.doExport(fastaPath, seqDict, fmt="fasta")
            ok = mU.doExport(taxonPath, taxonL, fmt="list")
        except Exception as e:
            logger.exception("Failing %r with %s", fastaPath, str(e))
        return ok
